# -*- coding: utf-8 -*-
#
import inspect
from asyncio import iscoroutinefunction
from sanic.views import HTTPMethodView
from sanic.response import BaseHTTPResponse
from sanic.constants import HTTP_METHODS

from .model import ModelBase

from .utils import unpack, best_match_accept_mimetype


class MethodViewExt(HTTPMethodView):
    methods = None
    method_has_context = None

    @classmethod
    def as_view_named(cls, endpoint_name, *class_args, **class_kwargs):
        """Return view function for use with the routing system, that
        dispatches request to appropriate handler method.
        """

        view = super(MethodViewExt, cls).as_view(*class_args, **class_kwargs)
        view.__name__ = endpoint_name
        return view

class ResourceMeta(type):
    def __new__(mcs, name, bases, d):
        p_type = type.__new__(mcs, name, bases, d)
        if 'methods' not in d:
            methods = set(p_type.methods or [])
            method_has_context = p_type.method_has_context or {}
            for m in HTTP_METHODS:
                ml = m.lower()
                func = d.get(ml, None)
                if func:
                    methods.add(m)
                    s = inspect.signature(func)
                    if len(s.parameters) < 3:
                        continue
                    # We have more than just 'self' and 'request'
                    p = iter(s.parameters.items())
                    next(p)  # self/cls
                    next(p)  # request
                    p = list(p)
                    for (i, (k, v)) in enumerate(p):
                        if v.name == "context":
                            if v.default == v.empty:
                                method_has_context[m] = i+2
                            else:
                                method_has_context[m] = 'k'
                            continue


            # If we have no method at all in there we don't want to
            # add a method list.  (This is for instance the case for
            # the base class or another subclass of a base method view
            # that does not introduce new methods).
            if methods:
                p_type.methods = sorted(methods)
            p_type.method_has_context = method_has_context
        return p_type


class Resource(MethodViewExt, metaclass=ResourceMeta):
    """
    Represents an abstract sanic_restplus.Resource.

    Concrete resources should extend from this class
    and expose methods for each supported HTTP method.
    If a resource is invoked with an unsupported HTTP method,
    the API will return a response with status 405 Method Not Allowed.
    Otherwise the appropriate method is called and passed all arguments
    from the url rule used when adding the resource to an Api instance.
    See :meth:`~sanic_restplus.Api.add_resource` for details.
    """

    representations = None
    method_decorators = []

    def __init__(self, api=None, *args, **kwargs):
        self.api = api

    async def dispatch_request(self, request, *args, **kwargs):
        context = kwargs.pop('context', None)
        has_context = bool(context)
        requestmethod = request.method
        meth = getattr(self, requestmethod.lower(), None)
        if meth is None and requestmethod == 'HEAD':
            meth = getattr(self, 'get', None)
            requestmethod = 'GET'
        elif meth is None and requestmethod == 'OPTIONS':
            meth = getattr(self, 'get', None)
            requestmethod = 'GET'
        assert meth is not None, 'Unimplemented method {0!r}'.format(requestmethod)
        method_has_context = self.method_has_context.get(requestmethod, False)
        for decorator in self.method_decorators:
            meth = decorator(meth)

        self.validate_payload(request, meth)
        if has_context and method_has_context:
            if method_has_context == 'k' or len(kwargs) > 0:
                kwargs.setdefault('context', context)
            else:
                pos = int(method_has_context) - 2  # skip self and request
                args = list(args)
                args.insert(pos, context)
        do_await = iscoroutinefunction(meth)
        resp = meth(request, *args, **kwargs)
        if do_await:
            resp = await resp
        resp_type = type(resp)
        if issubclass(resp_type, BaseHTTPResponse):
            return resp
        elif inspect.isawaitable(resp):
            # Still have a coroutine or awaitable even after waiting.
            # let the output handler handle it
            return resp

        representations = self.representations or {}

        mediatype = best_match_accept_mimetype(request, representations, default=None)
        if mediatype in representations:
             # resp might be a coroutine. Wait for it
             data, code, headers = unpack(resp)
             resp = representations[mediatype](data, code, headers)
             resp.headers['Content-Type'] = mediatype
             return resp

        return resp

    def __validate_payload(self, request, expect, collection=False):
        '''
        :param ModelBase expect: the expected model for the input payload
        :param bool collection: False if a single object of a resource is
        expected, True if a collection of objects of a resource is expected.
        '''
        # TODO: proper content negotiation
        data = request.json
        if collection:
            data = data if isinstance(data, list) else [data]
            for obj in data:
                expect.validate(obj, self.api.refresolver, self.api.format_checker)
        else:
            expect.validate(data, self.api.refresolver, self.api.format_checker)

    def validate_payload(self, request, func):
        '''Perform a payload validation on expected model if necessary'''
        if getattr(func, '__apidoc__', False) is not False:
            doc = func.__apidoc__
            validate = doc.get('validate', None)
            validate = validate if validate is not None else self.api._validate
            if validate:
                for expect in doc.get('expect', []):
                    # TODO: handle third party handlers
                    if isinstance(expect, list) and len(expect) == 1:
                        if isinstance(expect[0], ModelBase):
                            self.__validate_payload(request, expect[0], collection=True)
                    if isinstance(expect, ModelBase):
                        self.__validate_payload(request, expect, collection=False)



# -*- coding: utf-8 -*-
#
from asyncio import iscoroutinefunction
from datetime import timedelta
from functools import update_wrapper

from sanic.constants import HTTP_METHODS
from sanic.request import Request as sanic_request
from sanic.response import HTTPResponse
from sanic.views import HTTPMethodView

from sanic_restplus.utils import unpack


def crossdomain(origin=None, methods=None, headers=None, expose_headers=None,
                max_age=21600, attach_to_all=True,
                automatic_options=True, credentials=False):
    if methods is not None:
        m = list(sorted(x.upper() for x in methods))
        if 'OPTIONS' not in m:
            m.append('OPTIONS')
        methods = ', '.join(m)
    if headers is not None and not isinstance(headers, str):
        headers = ', '.join(x.upper() for x in headers)
    if expose_headers is not None and not isinstance(expose_headers, str):
        expose_headers = ', '.join(x.upper() for x in expose_headers)
    if not isinstance(origin, str):
        origin = ', '.join(origin)
    if isinstance(max_age, timedelta):
        max_age = max_age.total_seconds()

    def get_methods():
        if methods is not None:
            return methods
        # Todo:
        #  This is wrong for now, we need a way to find
        #  only the methods the httpmethodview contains
        return ', '.join(HTTP_METHODS)

    def decorator(f):
        async def wrapped_function(*args, **kwargs):
            orig_args = list(args)
            args_len = len(orig_args)
            rt = RuntimeError("Must only use crossdomain decorator on a function that takes 'request' as "
                              "first or second argument")
            if args_len < 1:
                #weird, no args
                raise rt
            elif args_len < 2:
                request = orig_args.pop(0)
                args = (request,)
            else:
                next_arg = orig_args.pop(0)
                args = list()
                if isinstance(next_arg, HTTPMethodView) or issubclass(next_arg, HTTPMethodView):
                    args.append(next_arg)  # self or cls
                    next_arg = orig_args.pop(0)
                request = next_arg
                args.append(request)
                args.extend(orig_args)
                args = tuple(args)
            if not isinstance(request, sanic_request):
                raise rt
            do_await = iscoroutinefunction(f)
            if automatic_options and request.method == 'OPTIONS':
                resp = HTTPResponse()
            else:
                resp = f(*args, **kwargs)
                if do_await:
                    resp = await resp
            if not attach_to_all and request.method != 'OPTIONS':
                return resp

            def apply_cors(h):
                nonlocal origin, get_methods, max_age, credentials, headers, expose_headers
                h['Access-Control-Allow-Origin'] = origin
                h['Access-Control-Allow-Methods'] = get_methods()
                h['Access-Control-Max-Age'] = str(max_age)
                if credentials:
                    h['Access-Control-Allow-Credentials'] = 'true'
                if headers is not None:
                    h['Access-Control-Allow-Headers'] = headers
                if expose_headers is not None:
                    h['Access-Control-Expose-Headers'] = expose_headers

            if isinstance(resp, HTTPResponse):
                apply_cors(resp.headers)
            elif isinstance(resp, tuple):
                resp, status, h = unpack(resp)
                apply_cors(h)
                resp = resp, status, h
            elif isinstance(resp, (str, list, dict, set, frozenset)):
                h = dict()
                apply_cors(h)
                resp = resp, 200, headers
            else:
                raise RuntimeError("crossorigin wrapper did not get a valid response from the wrapped function")
            return resp

        f.provide_automatic_options = False
        return update_wrapper(wrapped_function, f)
    return decorator



# -*- coding: utf-8 -*-
#
from sanic import exceptions

from ._http import HTTPStatus

__all__ = (
    'abort',
    'RestError',
    'ValidationError',
    'SpecsError',
)


def abort(code=HTTPStatus.INTERNAL_SERVER_ERROR, message=None, **kwargs):
    """
    Properly abort the current request.

    Raise a sanic exception for the corresponding error code

    :param int code: The associated HTTP status code
    :param str message: An optional details message
    :param kwargs: Any additional data to pass to the error payload  # TODO: ignored
    :return: Nothing, expect an exception raised
    :rtype: NoneType
    """
    status = int(code)
    return exceptions.abort(status_code=status, message=message)


class RestError(Exception):
    """Base class for all Flask-Restplus Errors"""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class ValidationError(RestError):
    """An helper class for validation errors."""
    pass


class SpecsError(RestError):
    """An helper class for incoherent specifications."""
    pass



# -*- coding: utf-8 -*-
#
import logging
import re
from collections import OrderedDict
from inspect import isclass

from .errors import RestError

log = logging.getLogger(__name__)

LEXER = re.compile(r'\{|\}|\,|[\w_:\-\*]+')


class MaskError(RestError):
    '''Raised when an error occurs on mask'''
    pass


class ParseError(MaskError):
    '''Raised when the mask parsing failed'''
    pass


class Mask(OrderedDict):
    '''
    Hold a parsed mask.

    :param str|dict|Mask mask: A mask, parsed or not
    :param bool skip: If ``True``, missing fields won't appear in result
    '''
    def __init__(self, mask=None, skip=False, **kwargs):
        self.skip = skip
        if isinstance(mask, str):
            super(Mask, self).__init__()
            self.parse(mask)
        elif isinstance(mask, (dict, OrderedDict)):
            super(Mask, self).__init__(mask, **kwargs)
        else:
            self.skip = skip
            super(Mask, self).__init__(**kwargs)

    def parse(self, mask):
        '''
        Parse a fields mask.
        Expect something in the form::

            {field,nested{nested_field,another},last}

        External brackets are optionals so it can also be written::

            field,nested{nested_field,another},last

        All extras characters will be ignored.

        :param str mask: the mask string to parse
        :raises ParseError: when a mask is unparseable/invalid

        '''
        if not mask:
            return

        mask = self.clean(mask)
        fields = self
        previous = None
        stack = []

        for token in LEXER.findall(mask):
            if token == '{':
                if previous not in fields:
                    raise ParseError('Unexpected opening bracket')
                fields[previous] = Mask(skip=self.skip)
                stack.append(fields)
                fields = fields[previous]
            elif token == '}':
                if not stack:
                    raise ParseError('Unexpected closing bracket')
                fields = stack.pop()
            elif token == ',':
                if previous in (',', '{', None):
                    raise ParseError('Unexpected comma')
            else:
                fields[token] = True

            previous = token

        if stack:
            raise ParseError('Missing closing bracket')

    def clean(self, mask):
        '''Remove unecessary characters'''
        mask = mask.replace('\n', '').strip()
        # External brackets are optional
        if mask[0] == '{':
            if mask[-1] != '}':
                raise ParseError('Missing closing bracket')
            mask = mask[1:-1]
        return mask

    def apply(self, data):
        '''
        Apply a fields mask to the data.

        :param data: The data or model to apply mask on
        :raises MaskError: when unable to apply the mask

        '''
        from . import fields
        # Should handle lists
        if isinstance(data, (list, tuple, set)):
            return [self.apply(d) for d in data]
        elif isinstance(data, (fields.Nested, fields.List, fields.Polymorph)):
            return data.clone(self)
        elif type(data) == fields.Raw:
            return fields.Raw(default=data.default, attribute=data.attribute, mask=self)
        elif data == fields.Raw:
            return fields.Raw(mask=self)
        elif isinstance(data, fields.Raw) or isclass(data) and issubclass(data, fields.Raw):
            # Not possible to apply a mask on these remaining fields types
            raise MaskError('Mask is inconsistent with model')
        # Should handle objects
        elif (not isinstance(data, (dict, OrderedDict)) and hasattr(data, '__dict__')):
            data = data.__dict__

        return self.filter_data(data)

    def filter_data(self, data):
        '''
        Handle the data filtering given a parsed mask

        :param dict data: the raw data to filter
        :param list mask: a parsed mask tofilter against
        :param bool skip: whether or not to skip missing fields

        '''
        out = {}
        for field, content in self.items():
            if field == '*':
                continue
            elif isinstance(content, Mask):
                nested = data.get(field, None)
                if self.skip and nested is None:
                    continue
                elif nested is None:
                    out[field] = None
                else:
                    out[field] = content.apply(nested)
            elif self.skip and field not in data:
                continue
            else:
                out[field] = data.get(field, None)

        if '*' in self.keys():
            for key, value in data.items():
                if key not in out:
                    out[key] = value
        return out

    def __str__(self):
        return '{{{0}}}'.format(','.join([
            ''.join((k, str(v))) if isinstance(v, Mask) else k
            for k, v in self.items()
        ]))


def apply(data, mask, skip=False):
    '''
    Apply a fields mask to the data.

    :param data: The data or model to apply mask on
    :param str|Mask mask: the mask (parsed or not) to apply on data
    :param bool skip: If rue, missing field won't appear in result
    :raises MaskError: when unable to apply the mask

    '''
    return Mask(mask, skip).apply(data)



# -*- coding: utf-8 -*-
#
import decimal
from sanic import exceptions
from collections import Hashable
from copy import deepcopy

from .errors import abort, SpecsError
from .marshalling import marshal
from .model import Model
from ._http import HTTPStatus

try:
    from sanic.compat import CIMultiDict
except ImportError:
    try:
        from sanic.server import CIMultiDict
    except ImportError:
        from sanic.server import CIDict as CIMultiDict

class ParseResult(dict):
    '''
    The default result container as an Object dict.
    '''
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value


_friendly_location = {
    'json': 'the JSON body',
    'form': 'the post body',
    'args': 'the query string',
    'values': 'the post body or the query string',
    'headers': 'the HTTP headers',
    'cookies': 'the request\'s cookies',
    'files': 'an uploaded file',
}

#: Maps Flask-RESTPlus RequestParser locations to Swagger ones
LOCATIONS = {
    'args': 'query',
    'form': 'formData',
    'headers': 'header',
    'json': 'body',
    'values': 'query',
    'files': 'formData',
}

#: Maps Python primitives types to Swagger ones
PY_TYPES = {
    int: 'integer',
    str: 'string',
    bool: 'boolean',
    float: 'number',
    None: 'void'
}

SPLIT_CHAR = ','

class Argument(object):
    '''
    :param name: Either a name or a list of option strings, e.g. foo or -f, --foo.
    :param default: The value produced if the argument is absent from the request.
    :param dest: The name of the attribute to be added to the object
        returned by :meth:`~reqparse.RequestParser.parse_args()`.
    :param bool required: Whether or not the argument may be omitted (optionals only).
    :param string action: The basic type of action to be taken when this argument
        is encountered in the request. Valid options are "store" and "append".
    :param bool ignore: Whether to ignore cases where the argument fails type conversion
    :param type: The type to which the request argument should be converted.
        If a type raises an exception, the message in the error will be returned in the response.
        Defaults to :class:`unicode` in python2 and :class:`str` in python3.
    :param location: The attributes of the :class:`flask.Request` object
        to source the arguments from (ex: headers, args, etc.), can be an
        iterator. The last item listed takes precedence in the result set.
    :param choices: A container of the allowable values for the argument.
    :param help: A brief description of the argument, returned in the
        response when the argument is invalid. May optionally contain
        an "{error_msg}" interpolation token, which will be replaced with
        the text of the error raised by the type converter.
    :param bool case_sensitive: Whether argument values in the request are
        case sensitive or not (this will convert all values to lowercase)
    :param bool store_missing: Whether the arguments default value should
        be stored if the argument is missing from the request.
    :param bool trim: If enabled, trims whitespace around the argument.
    :param bool nullable: If enabled, allows null value in argument.
    '''

    def __init__(self, name, default=None, dest=None, required=False,
                 ignore=False, type=str, location=('json', 'args',),
                 choices=(), action='store', help=None, operators=('=',),
                 case_sensitive=True, store_missing=True, trim=False,
                 nullable=True):
        self.name = name
        self.default = default
        self.dest = dest
        self.required = required
        self.ignore = ignore
        self.location = location
        self.type = type
        self.choices = choices
        self.action = action
        self.help = help
        self.case_sensitive = case_sensitive
        self.operators = operators
        self.store_missing = store_missing
        self.trim = trim
        self.nullable = nullable

    def source(self, request):
        '''
        Pulls values off the request in the provided location
        :param request: The flask request object to parse arguments from
        '''
        if isinstance(self.location, str):
            value = getattr(request, self.location, CIMultiDict())
            if callable(value):
                value = value()
            if value is not None:
                return value
        else:
            values = CIMultiDict()
            for l in self.location:
                value = getattr(request, l, None)
                if callable(value):
                    value = value()
                if value is not None:
                    values.update(CIMultiDict([(k,a) for k,v in value.items() for a in v]))
            return values

        return CIMultiDict()

    def convert(self, value, op):
        # Don't cast None
        if value is None:
            if not self.nullable:
                raise ValueError('Must not be null!')
            return None

        elif isinstance(self.type, Model) and isinstance(value, dict):
            return marshal(value, self.type)

        # and check if we're expecting a filestorage and haven't overridden `type`
        # (required because the below instantiation isn't valid for FileStorage)

        #TODO: update this for Sanic filestorage handler
        #elif isinstance(value, FileStorage) and self.type == FileStorage:
        #    return value

        try:
            return self.type(value, self.name, op)
        except TypeError:
            try:
                if self.type is decimal.Decimal:
                    return self.type(str(value), self.name)
                else:
                    return self.type(value, self.name)
            except TypeError:
                return self.type(value)

    def handle_validation_error(self, error, bundle_errors):
        '''
        Called when an error is raised while parsing. Aborts the request
        with a 400 status and an error message

        :param error: the error that was raised
        :param bool bundle_errors: do not abort when first error occurs, return a
            dict with the name of the argument and the error message to be
            bundled
        '''
        error_str = str(error)
        error_msg = ' '.join([str(self.help), error_str]) if self.help else error_str
        errors = {self.name: error_msg}

        if bundle_errors:
            return ValueError(error), errors
        abort(HTTPStatus.BAD_REQUEST, 'Input payload validation failed', errors=errors)

    def parse(self, request, req_context, bundle_errors=False):
        '''
        Parses argument value(s) from the request, converting according to
        the argument's type.

        :param request: The flask request object to parse arguments from
        :param bool bundle_errors: do not abort when first error occurs, return a
            dict with the name of the argument and the error message to be
            bundled
        '''

        try:
            bundle_errors = request.app.config.get('BUNDLE_ERRORS', False) or bundle_errors
        except AttributeError:
            bundle_errors = bundle_errors

        source = self.source(request)

        results = []

        # Sentinels
        _not_found = False
        _found = True

        for operator in self.operators:
            name = self.name + operator.replace('=', '', 1)
            if name in source:
                # Account for MultiDict and regular dict
                if hasattr(source, 'getlist'):
                    values = source.getlist(name)
                else:
                    values = [source.get(name)]

                for value in values:
                    if hasattr(value, 'strip') and self.trim:
                        value = value.strip()
                    if hasattr(value, 'lower') and not self.case_sensitive:
                        value = value.lower()

                        if hasattr(self.choices, '__iter__'):
                            self.choices = [choice.lower() for choice in self.choices]

                    try:
                        if self.action == 'split':
                            value = [self.convert(v, operator) for v in value.split(SPLIT_CHAR)]
                        else:
                            value = self.convert(value, operator)
                    except Exception as error:
                        if self.ignore:
                            continue
                        return self.handle_validation_error(error, bundle_errors)

                    if self.choices and value not in self.choices:
                        msg = 'The value \'{0}\' is not a valid choice for \'{1}\'.'.format(value, name)
                        return self.handle_validation_error(msg, bundle_errors)

                    if name in req_context.unparsed_arguments:
                        req_context.unparsed_arguments.pop(name)
                    results.append(value)

        if not results and self.required:
            if isinstance(self.location, str):
                location = _friendly_location.get(self.location, self.location)
            else:
                locations = [_friendly_location.get(loc, loc) for loc in self.location]
                location = ' or '.join(locations)
            error_msg = 'Missing required parameter in {0}'.format(location)
            return self.handle_validation_error(error_msg, bundle_errors)

        if not results:
            if callable(self.default):
                return self.default(), _not_found
            else:
                return self.default, _not_found

        if self.action == 'append':
            return results, _found

        if self.action == 'store' or len(results) == 1:
            return results[0], _found
        return results, _found

    @property
    def __schema__(self):
        if self.location == 'cookie':
            return
        param = {
            'name': self.name,
            'in': LOCATIONS.get(self.location, 'query')
        }
        _handle_arg_type(self, param)
        if self.required:
            param['required'] = True
        if self.help:
            param['description'] = self.help
        if self.default is not None:
            param['default'] = self.default() if callable(self.default) else self.default
        if self.action == 'append':
            param['items'] = {'type': param['type']}
            param['type'] = 'array'
            param['collectionFormat'] = 'multi'
        if self.action == 'split':
            param['items'] = {'type': param['type']}
            param['type'] = 'array'
            param['collectionFormat'] = 'csv'
        if self.choices:
            param['enum'] = self.choices
            param['collectionFormat'] = 'multi'
        return param


class RequestParser(object):
    '''
    Enables adding and parsing of multiple arguments in the context of a single request.
    Ex::

        from sanic_restplus import RequestParser

        parser = RequestParser()
        parser.add_argument('foo')
        parser.add_argument('int_bar', type=int)
        args = parser.parse_args()

    :param bool trim: If enabled, trims whitespace on all arguments in this parser
    :param bool bundle_errors: If enabled, do not abort when first error occurs,
        return a dict with the name of the argument and the error message to be
        bundled and return all validation errors
    '''

    def __init__(self, argument_class=Argument, result_class=ParseResult,
            trim=False, bundle_errors=False):
        self.args = []
        self.argument_class = argument_class
        self.result_class = result_class
        self.trim = trim
        self.bundle_errors = bundle_errors

    def add_argument(self, *args, **kwargs):
        '''
        Adds an argument to be parsed.

        Accepts either a single instance of Argument or arguments to be passed
        into :class:`Argument`'s constructor.

        See :class:`Argument`'s constructor for documentation on the available options.
        '''

        if len(args) == 1 and isinstance(args[0], self.argument_class):
            self.args.append(args[0])
        else:
            self.args.append(self.argument_class(*args, **kwargs))

        # Do not know what other argument classes are out there
        if self.trim and self.argument_class is Argument:
            # enable trim for appended element
            self.args[-1].trim = kwargs.get('trim', self.trim)

        return self

    def parse_args(self, req, req_context, strict=False):
        '''
        Parse all arguments from the provided request and return the results as a ParseResult

        :param bool strict: if req includes args not in parser, throw 400 BadRequest exception
        :return: the parsed results as :class:`ParseResult` (or any class defined as :attr:`result_class`)
        :rtype: ParseResult
        '''
        result = self.result_class()

        # A record of arguments not yet parsed; as each is found
        # among self.args, it will be popped out
        unparsed_arguments = dict(self.argument_class('').source(req)) if strict else {}
        req_context['unparsed_arguments'] = unparsed_arguments
        errors = {}
        for arg in self.args:
            value, found = arg.parse(req, req_context, self.bundle_errors)
            if isinstance(value, ValueError):
                errors.update(found)
                found = None
            if found or arg.store_missing:
                result[arg.dest or arg.name] = value
        if errors:
            abort(HTTPStatus.BAD_REQUEST, 'Input payload validation failed', errors=errors)

        if strict and unparsed_arguments:
            arguments = ', '.join(unparsed_arguments.keys())
            msg = 'Unknown arguments: {0}'.format(arguments)
            raise exceptions.SanicException("Bad Request: {}".format(msg), status_code=405)
        return result

    def copy(self):
        '''Creates a copy of this RequestParser with the same set of arguments'''
        parser_copy = self.__class__(self.argument_class, self.result_class)
        parser_copy.args = deepcopy(self.args)
        parser_copy.trim = self.trim
        parser_copy.bundle_errors = self.bundle_errors
        return parser_copy

    def replace_argument(self, name, *args, **kwargs):
        '''Replace the argument matching the given name with a new version.'''
        new_arg = self.argument_class(name, *args, **kwargs)
        for index, arg in enumerate(self.args[:]):
            if new_arg.name == arg.name:
                del self.args[index]
                self.args.append(new_arg)
                break
        return self

    def remove_argument(self, name):
        '''Remove the argument matching the given name.'''
        for index, arg in enumerate(self.args[:]):
            if name == arg.name:
                del self.args[index]
                break
        return self

    @property
    def __schema__(self):
        params = []
        locations = set()
        for arg in self.args:
            param = arg.__schema__
            if param:
                params.append(param)
                locations.add(param['in'])
        if 'body' in locations and 'formData' in locations:
            raise SpecsError("Can't use formData and body at the same time")
        return params


def _handle_arg_type(arg, param):
    if isinstance(arg.type, Hashable) and arg.type in PY_TYPES:
        param['type'] = PY_TYPES[arg.type]
    elif hasattr(arg.type, '__apidoc__'):
        param['type'] = arg.type.__apidoc__['name']
        param['in'] = 'body'
    elif hasattr(arg.type, '__schema__'):
        param.update(arg.type.__schema__)
    elif arg.location == 'files':
        param['type'] = 'file'
    else:
        param['type'] = 'string'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# flake8: noqa

import io
import os
import re
import sys

from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install

RE_REQUIREMENT = re.compile(r'^\s*-r\s*(?P<filename>.*)$')

PYPI_RST_FILTERS = (
    # Replace Python crossreferences by simple monospace
    (r':(?:class|func|meth|mod|attr|obj|exc|data|const):`~(?:\w+\.)*(\w+)`', r'``\1``'),
    (r':(?:class|func|meth|mod|attr|obj|exc|data|const):`([^`]+)`', r'``\1``'),
    # replace doc references
    (r':doc:`(.+) <(.*)>`', r'`\1 <http://flask-restplus.readthedocs.org/en/stable\2.html>`_'),
    # replace issues references
    (r':issue:`(.+?)`', r'`#\1 <https://github.com/noirbizarre/flask-restplus/issues/\1>`_'),
    # Drop unrecognized currentmodule
    (r'\.\. currentmodule:: .*', ''),
)



class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        # PUT YOUR POST-INSTALL SCRIPT HERE or CALL A FUNCTION

        develop.run(self)


def rst(filename):
    '''
    Load rst file and sanitize it for PyPI.
    Remove unsupported github tags:
     - code-block directive
     - all badges
    '''
    content = io.open(filename).read()
    for regex, replacement in PYPI_RST_FILTERS:
        content = re.sub(regex, replacement, content)
    return content



def pip(filename):
    '''Parse pip reqs file and transform it to setuptools requirements.'''
    requirements = []
    for line in io.open(os.path.join('requirements', '{0}.pip'.format(filename))):
        line = line.strip()
        if not line or '://' in line or line.startswith('#'):
            continue
        requirements.append(line)
    return requirements


long_description = '\n'.join((
    rst('README.rst'),
    rst('CHANGELOG.rst'),
    ''
))


exec(compile(open('sanic_restplus/__about__.py', encoding="latin-1").read(), 'sanic_restplus/__about__.py', 'exec'))

install_requires = pip('install')
if sys.version_info < (3, 5):
    raise RuntimeError("Cannot install on Python version < 3.5")
doc_require = pip('doc')
tests_require = pip('test')

setup(
    name='sanic-restplus',
    version=__version__,
    description=__description__,
    long_description=long_description,
    url='https://github.com/ashleysommer/sanic-restplus',
    author='Ashley Sommer',
    author_email='ashleysommer@gmail.com',
    packages=find_packages(exclude=['tests', 'tests.*']),
    entry_points={
        'sanic_plugins':
            ['RestPlus = sanic_restplus.restplus:instance']
    },
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
        'doc': doc_require,
    },
    cmdclass={
        'develop': PostDevelopCommand,
    },
    license='MIT',
    use_2to3=False,
    zip_safe=False,
    keywords='sanic restplus rest api swagger openapi',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: System :: Software Distribution',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ],
)



# -*- coding: utf-8 -*-
#
import re
from collections import OrderedDict
from copy import deepcopy
from ._http import HTTPStatus

#copied from sanic router
REGEX_TYPES = {
    'string': (str, r'[^/]+'),
    'int': (int, r'\d+'),
    'number': (float, r'[0-9\\.]+'),
    'alpha': (str, r'[A-Za-z]+'),
}

FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')


__all__ = ('merge', 'camel_to_dash', 'default_id', 'not_none', 'not_none_sorted', 'unpack')


def merge(first, second, _recurse=0):
    """
    Recursively merges two dictionaries.

    Second dictionary values will take precedence over those from the first one.
    Nested dictionaries are merged too.

    :param dict first: The first dictionary
    :param dict second: The second dictionary
    :return: the resulting merged dictionary
    :rtype: dict
    """
    if not isinstance(second, dict):
        return second
    result = deepcopy(first)
    for key, value in second.items():
        if key in result and isinstance(result[key], dict):
            if _recurse > 10:  # Max 10 dicts deep
                result[key] = None
            else:
                result[key] = merge(result[key], value, _recurse=_recurse+1)
        else:
            result[key] = deepcopy(value)
    return result


def camel_to_dash(value):
    '''
    Transform a CamelCase string into a low_dashed one

    :param str value: a CamelCase string to transform
    :return: the low_dashed string
    :rtype: str
    '''
    first_cap = FIRST_CAP_RE.sub(r'\1_\2', value)
    return ALL_CAP_RE.sub(r'\1_\2', first_cap).lower()


def default_id(resource, method):
    '''Default operation ID generator'''
    return '{0}_{1}'.format(method, camel_to_dash(resource))


def not_none(data):
    '''
    Remove all keys where value is None

    :param dict data: A dictionary with potentially some values set to None
    :return: The same dictionary without the keys with values to ``None``
    :rtype: dict
    '''
    return dict((k, v) for k, v in data.items() if v is not None)


def not_none_sorted(data):
    '''
    Remove all keys where value is None

    :param OrderedDict data: A dictionary with potentially some values set to None
    :return: The same dictionary without the keys with values to ``None``
    :rtype: OrderedDict
    '''
    return OrderedDict((k, v) for k, v in sorted(data.items()) if v is not None)


def unpack(response, default_code=HTTPStatus.OK):
    '''
    Unpack a Flask standard response.

    Flask response can be:
    - a single value
    - a 2-tuple ``(value, code)``
    - a 3-tuple ``(value, code, headers)``

    .. warning::

        When using this function, you must ensure that the tuple is not the response data.
        To do so, prefer returning list instead of tuple for listings.

    :param response: A Flask style response
    :param int default_code: The HTTP code to use as default if none is provided
    :return: a 3-tuple ``(data, code, headers)``
    :rtype: tuple
    :raise ValueError: if the response does not have one of the expected format
    '''
    if not isinstance(response, tuple):
        # data only
        return response, default_code, {}
    elif len(response) == 1:
        # data only as tuple
        return response[0], default_code, {}
    elif len(response) == 2:
        # data and code
        data, code = response
        return data, code, {}
    elif len(response) == 3:
        # data, code and headers
        data, code, headers = response
        return data, code or default_code, headers
    else:
        raise ValueError('Too many response values')

# Shamelessly copied from werkzeug:
# https://github.com/pallets/werkzeug/blob/master/src/werkzeug/http.py
# for explanation of "media-range", etc. see Sections 5.3.{1,2} of RFC 7231
_accept_re = re.compile(
    r"""
    (                       # media-range capturing-parenthesis
      [^\s;,]+              # type/subtype
      (?:[ \t]*;[ \t]*      # ";"
        (?:                 # parameter non-capturing-parenthesis
          [^\s;,q][^\s;,]*  # token that doesn't start with "q"
        |                   # or
          q[^\s;,=][^\s;,]* # token that is more than just "q"
        )
      )*                    # zero or more parameters
    )                       # end of media-range
    (?:[ \t]*;[ \t]*q=      # weight is a "q" parameter
      (\d*(?:\.\d+)?)       # qvalue capturing-parentheses
      [^,]*                 # "extension" accept params: who cares?
    )?                      # accept params are optional
    """,
    re.VERBOSE,
)

def parse_accept_header(value):
    """Parses an HTTP Accept-* header.  This does not implement a complete
    valid algorithm but one that supports at least value and quality extraction.
    :param value: the accept header string to be parsed.
    :return: a list of ``(value, quality)`` tuples.
    """
    result = []
    for match in _accept_re.finditer(value):
        quality = match.group(2)
        if not quality:
            quality = 1
        else:
            quality = max(min(float(quality), 1), 0)
        result.append((match.group(1), quality))
    return result

def get_accept_mimetypes(request):
    accept_types = request.headers.get('accept', None)
    if accept_types is None:
        return {}
    # keep the order they appear!
    return OrderedDict([((s, q), s) for s, q in parse_accept_header(accept_types)])

def best_match_accept_mimetype(request, representations, default=None):
    if representations is None or len(representations) < 1:
        return default
    try:
        accept_mimetypes = get_accept_mimetypes(request)
        if accept_mimetypes is None or len(accept_mimetypes) < 1:
            return default
        # find exact matches, in the order they appear in the `Accept:` header
        found = []
        for accept_type, qual in accept_mimetypes:
            if accept_type in representations:
                found.append((qual, accept_type))
        # match special types, like "application/json;charset=utf8" where the first half matches.
        for accept_type, qual in accept_mimetypes:
            type_part = str(accept_type).split(';', 1)[0]
            if type_part in representations:
                t = (qual, type_part)
                if t not in found:
                    found.append(t)
        if len(found) < 1:
            # if _none_ of those match, then fallback to wildcard matching
            for accept_type, qual in accept_mimetypes:
                if accept_type == "*" or accept_type == "*/*" or accept_type == "*.*":
                    found.append((qual, default))
        qual, best = sorted(found, reverse=True)[0]
        if qual <= 0.0:
            return None
        return best
    except (AttributeError, LookupError):
        pass  # fallback is to always return default
    # never return None here unless default=None
    return default

def parse_rule(parameter_string):
    """Parse a parameter string into its constituent name, type, and
    pattern

    For example:
    `parse_parameter_string('<param_one:[A-z]>')` ->
        ('param_one', str, '[A-z]')

    :param parameter_string: String to parse
    :return: tuple containing
        (parameter_name, parameter_type, parameter_pattern)
    """
    # We could receive NAME or NAME:PATTERN
    if str(parameter_string).startswith('/'):
        parameter_string = parameter_string[1:]
    parameter_string = str(parameter_string).strip('<>')
    name = parameter_string
    pattern = 'string'
    if ':' in parameter_string:
        name, pattern = parameter_string.split(':', 1)

    default = (str, pattern)
    # Pull from pre-configured types
    _type, pattern = REGEX_TYPES.get(pattern, default)

    return name, _type, pattern



from sanic import Sanic, Blueprint
from spf import SanicPlugin, SanicPluginsFramework
from spf.plugin import PluginRegistration, PluginAssociated


class RestPlusAssociated(PluginAssociated):

    def api(self, *args, api_class=None, **kwargs):
        (plug, reg) = self
        return plug.api(reg, *args, api_class=api_class, **kwargs)


class RestPlus(SanicPlugin):

    AssociatedTuple = RestPlusAssociated

    def __init__(self, *args, **kwargs):
        super(RestPlus, self).__init__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        self = super(RestPlus, cls).__new__(cls, *args, **kwargs)
        self.apis = set()
        return self

    def on_registered(self, context, reg, *args, **kwargs):
        try:
            apis = context.get('apis')
        except (LookupError, AttributeError):
            apis = None
        if apis is None:
            context['apis'] = apis = set()
        app = context.app
        try:
            ext = getattr(app, 'extensions', None)
            assert ext is not None
        except (AttributeError, AssertionError):
            setattr(app, 'extensions', dict())

    def api(self, reg, *args, api_class=None, **kwargs):
        from .api import Api
        if isinstance(reg, PluginAssociated):
            (pl, reg) = reg
            (spf, _, _) = reg
        elif isinstance(reg, PluginRegistration):
            (spf, _, _) = reg
        elif isinstance(reg, SanicPluginsFramework):
            spf = reg
            reg = self.find_plugin_registration(spf)
        elif isinstance(reg, Sanic):
            app = reg
            spf = SanicPluginsFramework(app)
            reg = self.find_plugin_registration(spf)
        else:
            raise RuntimeError("the 'reg' argument must be a SPF, an Association, Registration, or an App!")
        context = self.get_context_from_spf(reg)
        app = context.app
        assert isinstance(app, (Sanic, Blueprint))
        if api_class is None:
            if args and len(args) > 0 and isinstance(args[0], Api):
                args = list(args)
                api_class = args.pop(0)
            else:
                api_class = Api
        if isinstance(api_class, type):
            if app in args:
                # don't want app in here!
                args = list(args)
                args.remove(app)
            assert issubclass(api_class, Api)
            api = api_class(*args, **kwargs)
            kwargs = dict()
        else:
            api = api_class
        assert isinstance(api, Api)

        api.init_api(reg, **kwargs)
        context.apis.add(api)
        return api

    def register_api(self, reg, api, *args, **kwargs):
        return self.api(reg, api, *args, **kwargs)


restplus = instance = RestPlus()

__all__ = ["restplus"]
